import os
import sys
from bs4 import BeautifulSoup
import matplotlib.pyplot as plt
import numpy as np
from numpy.polynomial.polynomial import polyfit
import shutil
import re
import copy

def get_files_in_folder(folder_path):
	files = []
	contents = os.listdir(folder_path)
	for item in contents:
		item_path = folder_path + '/' + item
		if os.path.isfile(item_path):
			files.append(item)

	return files

def filename_no_ext(file_name):
	name_list = file_name.split('.')[:-1]

	fnl_name = ""
	rep = 0
	for item in name_list:
		if rep>0:
			fnl_name += '.'
		fnl_name += item
		rep +=1

	return fnl_name

def file_ext_from_name(file_name):
	return file_name.split('.')[-1]

def filepath_for_browser(path):
	abs_path = "file://" + os.getcwd() + '/' + path
	return abs_path

def page_pretty(page_path):
    reload(sys)
    sys.setdefaultencoding('utf8')

    with open(page_path, "rb") as f:
        html_string = f.read()
        html_string = BeautifulSoup(html_string, 'html.parser').prettify()
    
    with open(page_path, "wb") as f:
        html_string.encode('ascii', 'ignore').decode('ascii')
        f.write(html_string)

def default_folders_arrangement(filepath,file_name,folder_suffix):
    main_folder_path = "".join(filepath.split("/")[:-1])
    folder_name = main_folder_path + '/' + file_name + folder_suffix
    
    if not os.path.exists(folder_name):
        return
    else:
        contents = os.listdir(folder_name)
        for item in contents:
            if os.path.isfile(folder_name+'/'+item):
                os.remove(folder_name+'/'+item)
        ####### relocating files folder ##########
        for item in os.listdir(folder_name):
            shutil.move(folder_name+'/'+item,main_folder_path+'/'+item)

def plot_graph(xvals,yvals,title,xlabel,ylabel,figname,dest_dir_path):
	fig_path = dest_dir_path+'/'+figname
	plt.plot(xvals,yvals)
	plt.title(title)
	plt.xlabel(xlabel)
	plt.ylabel(ylabel)
	plt.savefig(fname=fig_path)
	plt.clf()

def plot_mult_lines(xvals,yvals,colors,markers,labels,title,xlabel,ylabel,figname,dest_dir_path):
	fig_path = dest_dir_path+'/'+figname

	for x,y,c,m,l in zip(xvals,yvals,colors,markers,labels):
		plt.plot(x,y,color=c,marker=m,label=l)

	plt.legend()
	plt.title(title)
	plt.xlabel(xlabel)
	plt.ylabel(ylabel)
	plt.savefig(fname=fig_path)
	plt.clf()

def plot_bar_chart(xvals,yvals,title,xlabel,ylabel,figname,dest_dir_path):
	fig_path = dest_dir_path+'/'+figname
	y_pos = np.arange(len(xvals))
	plt.bar(x=y_pos, height=yvals)
	plt.gca().yaxis.grid(True)
	plt.xticks(y_pos,xvals)
	plt.title(title)
	plt.xlabel(xlabel)
	plt.ylabel(ylabel)
	plt.savefig(fname=fig_path)
	plt.clf()

def line_of_best_fit(xvals,yvals,title,xlabel,ylabel,figname,dest_dir_path):
	b, m = polyfit(xvals, yvals, 1)
	plt.plot(xvals, yvals, '.')
	yvals_bf = b + m * np.array(xvals)
	plt.plot(xvals, yvals_bf, '-',label='y = {} * x + {}'.format(m,b))
	plt.legend()

	fig_path = dest_dir_path+'/'+figname
	plt.title(title)
	plt.xlabel(xlabel)
	plt.ylabel(ylabel)
	plt.savefig(fname=fig_path)
	plt.clf()

def plot_cdf(plot_data_list,title,xlabel,ylabel,figname,dest_dir_path,stats_file):
	f = open(stats_file,"w+")
	for item in plot_data_list:
		sorted_points = np.sort(item['datapoints'])
		y_points = 1. * np.arange(len(sorted_points))/(len(sorted_points))
		plt.plot(sorted_points,y_points,color=item['color'],label=item['label'])
		f.write("{} -> LowerQuart: {} Median: {} UpperQuart: {}\n".format(item['label'],np.percentile(item['datapoints'], 25),np.percentile(item['datapoints'], 50),np.percentile(item['datapoints'], 75)))
	plt.legend()
	f.close()

	fig_path = dest_dir_path+'/'+figname
	plt.title(title)
	plt.xlabel(xlabel)
	plt.ylabel(ylabel)
	plt.savefig(fname=fig_path)
	plt.clf()

def sort_nicely(in_list):
	""" Sort the given list in the way that humans expect.
	"""
	convert = lambda text: int(text) if text.isdigit() else text
	alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ]
	in_list.sort(key=alphanum_key)


import string
import os
from bs4 import BeautifulSoup
import shutil
from compress import compress_image_by_percentage as compress
import copy
from PIL import Image
import numpy as np
from get_object_data import get_page_img_dims as getImageDimensions
from skimage import measure
import cv2

temp_folder = "tmp"
shouldReplace = False

def annotate_images(webpageFolder):
    images = {}
    ids =  list(string.ascii_uppercase+string.ascii_lowercase)
    webpage = filter(lambda x: ".html" in x,os.listdir(webpageFolder))
    assert(len(webpage) ==1)
    webpage = webpage[0]
    i = 0
    html_code = ""

    # annotate eah image with a unique ID
    with open(os.path.join(webpageFolder,webpage), "r") as f:
        soup = BeautifulSoup(f.read(),"lxml")
        for each_img in soup.findAll("img"):
            each_img['id'] = ids[i]
            i += 1
            images[each_img['id']] = each_img['src']
        html_code = soup.prettify("utf-8")
    
    # save webpage with updated image ID's
    with open(webpage,"w") as f:
        f.write(html_code)
    return images

def apply_transformations(transformations,webpageFolder,images):
    # Create a tmp fodler and copy original HTML and images to that folder
    os.mkdir(temp_folder)
    webpage = filter(lambda x: ".html" in x,os.listdir(webpageFolder))[0]
    webpage_comple = os.path.join(webpageFolder,webpage)
    shutil.copy(webpage_comple,os.path.join(temp_folder,"original.html"))

    new_images = {}
    for id in images:
        src = images[id]
        image_name = src.split("/")[-1]
        if image_name == "":
            image_name = src.split("/")[-2]
        new_images[id] = image_name
        shutil.copy(src, os.path.join(temp_folder,image_name))
    
    #Make all transformed Images in tmp folder
    images_names = filter(lambda x: x.split(".")[1] != "html" ,os.listdir(temp_folder))
    for eachImage in images_names:
        for percReduction in transformations["compress"]:
            newName = eachImage.split(".")[0] + "_"+str(percReduction) +"."+ eachImage.split(".")[-1]
            compress(temp_folder,temp_folder,eachImage,newName,percReduction)
    

    # Make a dict of all transformations
    folders = {}
    folders["original"] = new_images.keys()

    for id in images:
        for percReduction in transformations["compress"]:
            fname = "compress_" + id + "_" + str(percReduction)
            if fname not in folders.keys():
                folders[fname] = []
            newName = images[id].split("/")[-1]
            if newName == "":
                newName = images[id].split("/")[-2]
            newName_src = newName.split(".")[0] + "_"+str(percReduction) +"."+  images[id].split(".")[-1]
            newName_id  = id + "_" + str(percReduction) 
            new_images[newName_id] = newName_src
            
            folders[fname].append(newName_id)
            for j in images:
                if j != id:
                    folders[fname].append(j)
    
    if transformations["removal"]:
        #Generate White Images corresponding to original Images
        for id in images:
            generateWhiteImage(images[id],temp_folder,images[id].split("/")[-1].split(".")[0])
        
        for id in images:
            name = "remove_" + id
            folders[name] = [name]
            new_images[name] =  images[id].split("/")[-1].split(".")[0]+"_removal.jpg"
            for otherImage in images:
                if otherImage != id:
                    folders[name].append(otherImage)


    #Create folders Now
    for eachFolder in folders:
        if not os.path.exists(os.path.join(temp_folder,eachFolder)):
            os.mkdir(os.path.join(temp_folder,eachFolder))
        for eachFile in folders[eachFolder]:
            shutil.copy( os.path.join(temp_folder,new_images[eachFile]), os.path.join(temp_folder, eachFolder,new_images[eachFile]))
        
    # Create HTML files
    for eachFolder in folders:
        with open(os.path.join(temp_folder,"original.html"), "r") as f:
            soup = BeautifulSoup(f.read(),"lxml")
            for each_img in soup.findAll("img"):
                img_id = filter(lambda x: each_img["id"] + "_" in x or "_"+each_img["id"] in x or each_img["id"] == x , folders[eachFolder])[0]
                each_img['src'] = new_images[img_id]
            html_code = soup.prettify("utf-8")
            with open(os.path.join(temp_folder,eachFolder,eachFolder+".html"), "w") as new_f:
                new_f.write(html_code)

    return new_images,folders

def calculate_ssim(images,transformed_webpages):
    final_data = {}
    for eachFile in transformed_webpages:
        data = getImageDimensions(os.path.join(temp_folder,eachFile,eachFile+".html"),eachFile+".html")
        final_data[eachFile] = data
    
    final_data = add_ssim_to_data(final_data)

    for eachPage in final_data:
        final_data[eachPage]["WMSSIM"] =  calculateWeightedMeanSSIM(final_data[eachPage])
    clear_tmp_folder()
    return final_data


def generateWhiteImage(image, imagesFolder,id):
    with Image.open(image) as img:
        width, height = img.size

    img = np.zeros([height,width,3],dtype=np.uint8)
    img.fill(255)
    im = Image.fromarray(img)
    im.save(os.path.join(imagesFolder,id+"_removal.jpg"))
    return

def clear_tmp_folder():
    #Delete tmp Folder
    shutil.rmtree(temp_folder)


def add_ssim_to_data(imagesData):
    basePage = "original"
    baseImages = imagesData[basePage]['imgs'].keys()
    baseImages.sort()

    pages = filter(lambda x: x != basePage, imagesData.keys())

    for eachImage in baseImages:
        imagesData[basePage]['imgs'][eachImage]['SSIM'] = 1
    for eachPage in pages:
        for eachImage in baseImages:
            pageImage = filter(lambda x:  eachImage+ "_" in x or "_"+eachImage in x or eachImage == x,  imagesData[eachPage]['imgs'].keys())[0]
            if "remove_" in pageImage and not shouldReplace:
                imagesData[eachPage]['imgs'][pageImage]['SSIM']  = 0
            elif imagesData[eachPage]['imgs'][pageImage]['src'] == imagesData[basePage]['imgs'][eachImage]['src']:
                imagesData[eachPage]['imgs'][pageImage]['SSIM']  = 1
            else:
                imagesData[eachPage]['imgs'][pageImage]['SSIM'] = find_ssim(imagesData[eachPage]['imgs'][pageImage]['src'], imagesData[basePage]['imgs'][eachImage]['src'], temp_folder)
    return imagesData

def find_ssim(img1, img2, filePath):
    return get_ssim_value(os.path.join(filePath,img1),os.path.join(filePath,img2))

def get_ssim_value(imageA,imageB):
    imageA = cv2.imread(imageA)
    imageB = cv2.imread(imageB)
    H,W,A = imageA.shape
    imageB = cv2.resize(imageB,(W,H))
    s = measure.compare_ssim(imageA,imageB,multichannel=True)
    return s


def calculateWeightedMeanSSIM(pageData):
    WMSSIM = 0
    for eachImage in pageData['imgs'].keys():
        if 'SSIM' in pageData['imgs'][eachImage].keys():
            WMSSIM += weightOfImage(eachImage,pageData)*pageData['imgs'][eachImage]['SSIM']
        else:
            return -1
    return WMSSIM

def weightOfImage(image,pageData):
    return (locationWeightOfImage(image,pageData) + areaWeightOfImage(image,pageData))/2.0


# locW(img1) = dist1/(dist1+dist2+...)
def locationWeightOfImage(image,pageData):
    dist1 = float(pageData['body']['height'] - pageData['imgs'][image]['yloc'])

    totalDist = 0
    for eachImage in pageData['imgs'].keys():
        totalDist += float(pageData['body']['height'] - pageData['imgs'][eachImage]['yloc'])
    
    return dist1/totalDist

# areaW(img1) = area1/(area1+area2+...)
def areaWeightOfImage(image,pageData):
    area1 = float(pageData['imgs'][image]['width'] * pageData['imgs'][image]['height'])

    totalArea = 0
    for eachImage in pageData['imgs'].keys():
        totalArea += float(pageData['imgs'][eachImage]['width'] * pageData['imgs'][eachImage]['height'])
    
    return area1/totalArea


#This script returns a json file containing dimensions and location for all images in the web pages it
#is given. It also returns dimensions for the web page body

# output json format:  
# {filename1:
# 	{'imgs':
# 		{img1:{height:1,width:1,area:1,xloc:20,yloc:400},img2:{height:2,width:2,area:4,xloc:412,yloc:213},...}
# 	'body': {height:10,width:10,area:100}
# 	},
#  filename2: ...
# }

import sys
import util
from selenium import webdriver
import getopt
import os
import json
from pyvirtualdisplay import Display


def setup_driver():
    driver = webdriver.Firefox(executable_path='/home/ab/geckodriver')
    driver.maximize_window()
    return driver

def get_page_img_dims(filepath,filename_nx):
    with Display():
        sizes_dict = {}
        driver = setup_driver()
        abs_path = util.filepath_for_browser(filepath)
        sizes_dict[filename_nx] = {'imgs':{}}
        driver.get(abs_path)
        body_tag = driver.find_element_by_tag_name('body')
        img_tags = driver.find_elements_by_tag_name('img')

        body_dims = body_tag.size
        body_dims['area'] = body_tag.size['width'] * body_tag.size['height']
        sizes_dict[filename_nx]['body'] = body_dims
        for tag in img_tags:
            try:
                img_name = tag.get_attribute('src').split("/")[-1]
                if not img_name:
                    img_name = tag.get_attribute('src').split("/")[-2]
                img_name_nx = util.filename_no_ext(img_name)
                dims = tag.size
                dims['area'] = tag.size['width'] * tag.size['height']
                sizes_dict[filename_nx]['imgs'][img_name_nx] = dims
                sizes_dict[filename_nx]['imgs'][img_name_nx]['xloc'] = tag.location['x']
                sizes_dict[filename_nx]['imgs'][img_name_nx]['yloc'] = tag.location['y']
                sizes_dict[filename_nx]['imgs'][img_name_nx]['src'] = img_name

            except:
                driver.quit()
                continue

        driver.quit()
        import copy
        return copy.deepcopy(sizes_dict[filename_nx])



from helper_functions import annotate_images, apply_transformations, calculate_ssim

# transformations = { "compress": [25,50,75,90], 
#                     "removal" : True        }
# webpage = Folder address that must contain an HTML file
# weighted SSIM = ["imageID":{   
#                             "src": ".\link.jpg"
#                             "compress_25": 0.915,
#                             "compress_50": 0.80,
#                             "removal" : 0.0       }]
def get(webpage,transformations):
    try:
        images = annotate_images(webpage)
        new_images,transformed_webpages = apply_transformations(transformations,webpage,images)
        weighted_ssim = calculate_ssim(new_images,transformed_webpages)
        return beautify_result(images, transformations, weighted_ssim)
    except Exception as e:
        print("Error: Please make sure you are following the correct syntax.")
        print(e)

def beautify_result(images,transformations,data):
    # {
    #     ID: {
    #             src : '',
    #             compress_25 : 0.9,
    #             compress_50 : 0.8,
    #             remove      : 0.9,
    #         },
    # }

    result = {}
    for eachImage in images:
        result[eachImage] = {}
        result[eachImage]["src"] = images[eachImage]
        for compression in transformations["compress"]:
            result[eachImage]["compress_"+ str(compression)] = data["compress_"+eachImage+"_" + str(compression)]["WMSSIM"]
        if transformations["removal"]:
            result[eachImage]["remove"] = data["remove_"+eachImage]["WMSSIM"]
    return result


import os
import sys
from PIL import Image
import numpy
import magic
import shutil


#compresses by taking image quality
def compress_image(src_folder,target_folder,img_file,new_name,quality_metric):
	ext_dict = {"jpg":"JPEG","jpeg":"JPEG","png":"PNG","gif":"GIF"}
	file_info = magic.from_file(src_folder+'/'+img_file, mime=True)
	img_ext = file_info.split('/')[-1]

	#check if image format is recognized
	if img_ext not in ext_dict.keys():
		#print "Error: unrecognized image extension {} in file {}".format(img_ext,src_folder+'/'+img_file)
		return -1,-1,-1
	format = ext_dict[img_ext]

	#limiting image quality between 1 and 100. can be changes
	if not(1 <= quality_metric <= 100):
		print ("Error: Quality must be in the range {} to {}. Given value: {}".format(1,100,quality_metric))
		sys.exit(1)
	filepath = os.path.join(os.getcwd(), src_folder+"/"+img_file)
	oldsize = float(os.stat(filepath).st_size/1000.0)
	picture = Image.open(filepath)
	newpath = os.path.join(os.getcwd(),target_folder+"/"+new_name)
	picture.save(target_folder+"/"+new_name,format,optimize=True,quality=quality_metric) 
	newsize = float(os.stat(newpath).st_size/1000.0)
	percent = (oldsize-newsize)/float(oldsize)*100
	return oldsize,newsize,percent


#tells which value is in the 1 to 100 range
def val_in_range(low,high):
	val_range = range(1,101)
	
	if (high in val_range) and (low in val_range):
		return (low+high)/2
	elif high in val_range:
		return high
	elif low in val_range:
		return low
	else:
		print ("Error in val_in_range: low value {} and high value {} both out of range".format(low,high))
		sys.exit(1)

def size_comparison(newsize,org_size,percentage):
	slack = 1 #percent
	percent_upper = percentage+slack
	percent_lower = percentage-slack
	percent_achieved = (1 - (float(newsize)/float(org_size)))*100

	if percent_achieved<percent_lower: #not enough size reduction achieved
		return  1
	elif percent_achieved>percent_upper: #too much reduction achieved
		return -1
	else:
		return 0

def quality_binary_search(filepath,percentage,low,high,org_size,format,ext):
	#no ideal quality found. Returns either high or low(whichever is in the 1-100 range)
	if high<low:
		return val_in_range(low,high)

	mid = int((low+high)/2)
	new_filename = "size_test."+ext
	picture = Image.open(filepath)
	picture.save(new_filename,format,optimize=True,quality=mid)
	newsize = float(os.stat(new_filename).st_size/1000.0)
	#checking image size
	size_comp = size_comparison(newsize,org_size,percentage)
	#delete photo after checking size
	os.remove(new_filename)
	if size_comp==0: #size in range
		return mid
	elif size_comp==-1: #size too low
		return quality_binary_search(filepath,percentage,mid+1,high,org_size,format,ext)
	elif size_comp==1: #size too high
		return quality_binary_search(filepath,percentage,low,mid-1,org_size,format,ext)


#compresses by taking percentage reduction
def compress_image_by_percentage(src_folder,target_folder,img_file,new_name,percentage_reduction):
	ext_dict = {"jpg":"JPEG","jpeg":"JPEG","png":"PNG","gif":"GIF"}
	file_info = magic.from_file(src_folder+'/'+img_file, mime=True)
	img_ext = file_info.split('/')[-1]

	#check if image format is recognized
	if img_ext not in ext_dict.keys():
		print ("Error: unrecognized image extension {} in file {}".format(img_ext,src_folder+'/'+img_file))
		return -1,-1,-1,-1
	format = ext_dict[img_ext]

	#limiting percentage reduction between 0 and 100%. Can be extended
	if not(0 <= percentage_reduction <= 100):
		print ("Error: Percentage must be in the range {} to {}".format(0,100))
		sys.exit(1)

	filepath = os.path.join(os.getcwd(), src_folder+"/"+img_file)
	oldsize = float(os.stat(filepath).st_size/1000.0)

	#finding quality level to achieve desired percentage reduction
	quality_level = quality_binary_search(filepath,percentage_reduction,1,100,oldsize,format,img_ext)

	picture = Image.open(filepath)
	newpath = os.path.join(os.getcwd(),target_folder+"/"+new_name)
	picture.save(target_folder+"/"+new_name,format,optimize=True,quality=quality_level) 
	newsize = float(os.stat(newpath).st_size/1000.0)
	percent = (oldsize-newsize)/float(oldsize)*100

	#handle case in which compression increased image size
	if percent<0:
		os.remove(newpath)
		shutil.copyfile(filepath,newpath)
		return oldsize,oldsize,0,-1		

	return oldsize,newsize,percent,quality_level


if __name__=='__main__':
	src_folder = sys.argv[1]
	percentage_reduction = 25
	target_folder = src_folder+"_compressed_"+str(percentage_reduction)
	if not os.path.isdir(target_folder):
		os.mkdir(target_folder)
	images = os.listdir(src_folder)
	percentages = []
	for img in images:
		oldsize,newsize,percent,quality = compress_image_by_percentage(src_folder,target_folder,img,"Compressed_"+img,percentage_reduction)
		#percentages.append(percent)
		#print "Image {} compressed by {} percent from {} KB to {} KB ::: Quality Metric : {}".format(img,percent,oldsize,newsize,quality_metric)
		print ("Img {} -> Target Percentage: {} Percentage Achieved: {} Quality: {}".format(img,percentage_reduction,percent,quality))
	# avg = sum(percentages)/len(percentages)
	# print "Average compression for quality {} is {} percent".format(quality_metric,avg)
	# print "Standard deviation for quality {} is {} percent".format(quality_metric,numpy.std(percentages))




from setuptools import setup,find_packages



setup(name='wm_ssim',
      version='0.2.1',
      description='The library to find WM_SSIM of a webapge with all the transformations',
      packages = ['wm_ssim'],
      url='http://github.com/abdul-manaan/wm_ssim',
      author='Abdul Manan',
      author_email='manan.lums@gmail.com',
      license='MIT',
      install_requires=[
            'selenium==3.141.0',
            'numpy==1.11.0',
            'opencv_python==4.1.1.26',
            'scikit_image==0.15.0',
            'matplotlib',
            'Pillow==6.1.0',
            'beautifulsoup4==4.8.0',
            'python_magic==0.4.15',
            'pyvirtualdisplay==0.2.4',
            'skimage==0.0'],
      zip_safe=False,
      classifiers=[
        'Programming Language :: Python :: 2'
      ])